/**
* Zeigt durch Rückgabewert von getHumanDbmsLevels definierte Level zum Lösen an
*/
function HumanDbmsTaskUI(
    nodeContainer       // DOM element to occupy
) {
    nodeContainer.innerHTML = '';
    this._levelMgr = new Levelliste(nodeContainer);
    this._levelMgr.onchange = ((levelNr, oldLevelNr) => this._startLevel(levelNr, oldLevelNr));
    this._levelMgr.strLevelName = 'Aufgabe ';
    this._levelMgr.onnamelevel = (nr => (nr > this.task.levels.length) ? 'Abschluss' : '');
    this._showTask('Wird geladen…');
    this._displayTables = new DbTableDisplay(nodeContainer);
    this._displayTables.oncopy = ({ csv }) => this._editableTable.fromString(csv);
    this._msgPlace = new ScoTools.MessagePlace(false, true);
    this._msgPlace.appendAsChild(nodeContainer);
    this._nodeResultTable = ScoTools.createHtml('section', nodeContainer);
    ScoTools.createHtml('h2', this._nodeResultTable, null, 'Ergebnistabelle');
    this._editableTable = new EditableTable();
    this._editableTable.onchange = () => this._procTableChange();
    this._nodeResultTable.appendChild(this._editableTable.node);
    this._nextBtn = ScoTools.createHtml('button', nodeContainer, { 'class': 'fullwidth' });
    ScoTools.createHtml('img', this._nextBtn, { src: 'share/Go-next.svg', alt: '' });
    ScoTools.createHtml(false, this._nextBtn, null, ' Weiter');
    this._nextBtn.addEventListener('click', () => this.checkInput());
}

// Verwendete Levelliste-Instanz
HumanDbmsTaskUI.prototype._levelMgr = null;

// Container (DOM-Knoten) für Terminal, ggf. Testfrage und Weiter-Schaltfläche
HumanDbmsTaskUI.prototype._contentContainer = null;

// Container (DOM-Knoten) für Testfrage
HumanDbmsTaskUI.prototype._containerQuestion = null;

// DOM-Knoten für „Weiter“-Schaltfläche
HumanDbmsTaskUI.prototype._btnNext = null;

// Muss erst eine Tabelle versandt werden?
HumanDbmsTaskUI.prototype.locked = false;

// Funktion, die beim Freischalten eines neuen Levels aufgerufen werden soll
HumanDbmsTaskUI.prototype.onlevelunlock = function(unlockedLevelCount) { }

// Funktion, die bei einer Veränderung der eingegebenen Tabelle aufgerufen werden soll
HumanDbmsTaskUI.prototype.ontablechange = function(csv) { }

// Zahl der bereits freigeschalteten Level
Object.defineProperty(HumanDbmsTaskUI.prototype, 'unlockedLevelCount', {
    enumerable: true,
    get() { return Math.min(this._levelMgr.done, this.task.levels.length); },
    set(v) {
        this._levelMgr.done = v + ((this.task.levels.length == v) ? 1 : 0);
        this._levelMgr.active = v;
    }
});

// Wurden bereits alle Level frei geschaltet?
Object.defineProperty(HumanDbmsTaskUI.prototype, 'done', {
    enumerable: true,
    get() { return (this._levelMgr.done > this.task.levels.length); }
});

// Funktion, die beim Erhalt von Strafpunkten aufgerufen werden soll
HumanDbmsTaskUI.prototype.onnewfine = function(totalFine, additionalFine) { }

// Gesammelte Anzahl an Fehlerpunkten
HumanDbmsTaskUI.prototype._fineCount = 0;
Object.defineProperty(HumanDbmsTaskUI.prototype, 'fineCount', {
    enumerable: true,
    get() { return this._fineCount; },
    set(v) {
        if ((v == this._fineCount) || !Number.isFinite(v))
            return;
        const oldFine = this._fineCount;
        this._fineCount = v;
        if ((typeof this.onnewfine == 'function') || (this.onnewfine instanceof Function))
            this.onnewfine.call(this, v, v - oldFine);
    }
});

/** Startwert für Pseudozufallszahlengenerierung */
HumanDbmsTaskUI.prototype.startSeed = 1;

/** Gespeicherter Rückgabewert von getHumanDbmsLevels */
HumanDbmsTaskUI.prototype.tasks = { tabletype: 'Ergebnistabelle', tables: [], levels: [] };

/** Tabellentyp */
Object.defineProperty(HumanDbmsTaskUI.prototype, 'tabletype', {
    enumerable: true,
    get() { return (this.tasks.tabletype ? this.tasks.tabletype : 'Ergebnistabelle'); }
});

/**
* Früher gespeicherte Werte wieder übernehmen
* @return undefined
*/
HumanDbmsTaskUI.prototype.loadValues = function(
    seed,                   // Startwert für Pseudozufallszahlengenerierung
    unlockedLevelCount,     // Anzahl der freigeschalteten Level
    fineCount,              // Anzahl der gesammelten Fehlerpunkte
    csv                     // Tabelleninhalt als CSV
) {
    this.startSeed = seed;
console.log('seed:', this.startSeed);
    const fktRandomInt = ((min, max) => { seed = (seed * 9301 + 49297) % 233280; return Math.floor(seed / 233280 * (max - min + 1)) + min; });
    this.task = getHumanDbmsLevels(fktRandomInt);
    this._displayTables.tables = this.task.tables;
    this._fineCount = (Number.isFinite(fineCount) ? fineCount : 0);
    this._levelMgr.count = this.task.levels.length + 1;
    this.unlockedLevelCount = unlockedLevelCount;
    this._nodeResultTable.firstChild.firstChild.nodeValue = this.tabletype;
    if (csv)
        this._editableTable.fromString(csv);
}

/**
* Aufgabenstellung anzeigen
* @return undefined
*/
HumanDbmsTaskUI.prototype._showTask = function(
    task        // Anzuzeigender HTML-Code
) {
    if (!task && this._nodeTask) {
        this._nodeTask.parentNode.removeChild(this._nodeTask);
        this._nodeTask = null;
        return;
    }
    if (this._nodeTask) {
        this._nodeTask.innerHTML = task;
        return
    }
    this._nodeTask = ScoTools.createHtml('div', null, { 'class': 'auftrag' });
    this._nodeTask.innerHTML = task;
    this._levelMgr.node.parentNode.insertBefore(this._nodeTask, this._levelMgr.node.nextSibling);
}

/**
* Informationen über aktuelles Level ermitteln
* @return Object    Eintrag aus this.task.levels
*/
HumanDbmsTaskUI.prototype.getLevelInfo = function() {
    const level = this.task.levels[this._levelMgr.active];
    return level ? level : null;
}

/**
* Level (neu) starten
* @return undefined
*/
HumanDbmsTaskUI.prototype._startLevel = function(
    levelNr,        // Nummer des zu startenden Levels (beginnend bei 0)
    oldLevelNr      // Nummer des vorherigen Levels (beginnend bei 0)
) {
    const level = this.getLevelInfo();
    this.clearMsg();
    if (level) {
        this._nextBtn.style.display = '';
        this._nextBtn.lastChild.nodeValue = level.locked ? ' Tabelle überprüfen' : ' Weiter';
        this._displayTables.node.style.display = '';
        this._displayTables.showCopy = level.locked;

        if ((typeof level.start == 'function') || (level.start instanceof Function))
            level.start();
        this._showTask(level.task ? level.task : '');
        this._nodeResultTable.style.display = level.locked ? '' : 'none';
        if (level.locked)
            this._editableTable.fromString(level.placeholder ? level.placeholder : ';\n;');
    } else {
        this._nextBtn.style.display = 'none';
        this._nodeResultTable.style.display = 'none';
        this._displayTables.node.style.display = 'none';
        this._showTask(this.createFinalReport());
    }
}

/**
* Abschlussbericht zusammenstellen
* @return String        HTML-Code des Abschlussberichts
*/
HumanDbmsTaskUI.prototype.createFinalReport = function() {
    if (!this.fineCount)
        return 'Hervorragend! Du hast alle Aufgaben fehlerfrei gelöst.';
    if (this.fineCount == 1)
        return 'Gut gemacht: Du hast alle Aufgaben gelöst und nur <strong>einen Fehlerpunkt</strong> kassiert.';
    const arrZahlen = ['', '', 'zwei', 'drei', 'vier', 'fünf', 'sechs', 'sieben', 'acht', 'neun', 'zehn', 'elf', 'zwölf'];
    if (this.fineCount < arrZahlen.length)
        return 'Du hast alle Aufgaben gemeistert, dabei allerdings <strong>' + arrZahlen[this.fineCount] + ' Fehlerpunkte</strong> gesammelt.';
    return 'Du hast alle Aufgaben gemeistert, dabei allerdings <strong>' + this.fineCount + ' Fehlerpunkte</strong> gesammelt.';
}

// Anzuzeigende Meldung, wenn es sonst keine Rückmeldung gibt
HumanDbmsTaskUI.prototype._generalMsg = null;

/**
* Festlegen oder löschen, welche generische Meldung anzuzeigen ist, wenn sonst keine Meldung verfügbar ist (und diese gleich anzeigen, wenn eine Meldung festgelegt wird)
* @return undefined
*/
HumanDbmsTaskUI.prototype.setGeneralMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._generalMsg = text ? { text, type } : null;
    if (this._generalMsg)
        this.clearMsg();
}

/**
* Aktive Meldung löschen
* @return undefined
*/
HumanDbmsTaskUI.prototype.clearMsg = function() {
    if (this._generalMsg)
        this._msgPlace.add(this._generalMsg.text, this._generalMsg.type);
    else
        this._msgPlace.clear();
}

/**
* Eine Meldung vorübergehend anzeigen
* @return undefined
*/
HumanDbmsTaskUI.prototype.addMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._msgPlace.add(text, type);
}

/**
* Eingabe überprüfen und - sofern alles richtig ist - zum nächsten Level wechseln
* @return undefined
*/
HumanDbmsTaskUI.prototype.checkInput = function() {
    const level = this.getLevelInfo();
    if (!level.locked) {
        this.startNextLevel();
        return;
    }
    const data = this._editableTable.getValues();
    // Anhand vorgegebener Ergebnistabelle überprüfen
    if (level.result !== undefined) {
        const expected = EditableTable.stringToArray(level.result);
        // Attributnamen überprüfen
        if (expected[0].length != data[0].length)
            return this._tellUser('Statt ' + data[0].length + ' müssen es ' + expected[0].length + ' Attribute sein');
        let errorMsg = expected[0].reduce((prev, curr, index) => (prev || ((curr == data[0][index]) ? false : ('Der ' + (index + 1) + '. Attributname ist falsch: Statt „' + data[0][index] + '“ wäre „' + curr + '“ richtig'))), false);
        // Datensätze prüfen
        if (expected.length != data.length)
            return this._tellUser('Statt ' + (data.length - 1) + ' müssen es ' + (expected.length - 1) + ' Datensätze sein');
        data.splice(0, 1);
        expected.splice(0, 1);
        if (level.orderRelevant) {
            errorMsg = expected.reduce((prev, row, y) => row.reduce((p, value, x) => (p || (
                (value == data[y][x]) ? false : ('Beim ' + (y + 1) + '. Datensatz ist der ' + (x + 1) + '. Attributwert falsch: Statt „' + data[y][x] + '“ müsste es „' + value + '“ heißen')
            )), prev), errorMsg);
        } else {
            errorMsg = data.reduce((prev, rowInput, rowCount) => {
                if (prev)
                    return prev;
                const index = expected.reduce((found, rowSoll, index) => ((found == -1) ? (rowSoll.reduce((p, v, i) => (p && (v == rowInput[i])), true) ? index : -1) : found), -1);
                if (index == -1)
                    return 'Datensatz ' + (rowCount + 1) + ' ist falsch';
                expected.splice(index, 1);
                return false;
            }, errorMsg);
        }
        if (errorMsg)
            return this._tellUser(errorMsg);
    }
    // Anhand von Überpüfungsfunktion kontrollieren
    if ((typeof level.check == 'function') || (level.check instanceof Function)) {
        let checkResult = level.check.call(level, { data });
        if (checkResult)
            checkResult = (checkResult instanceof Object) ? checkResult : { unlock: true };
        else
            checkResult = { msg: 'Diese Tabelle ist falsch' };
        if (checkResult.msg)
            this._tellUser(checkResult.msg, (checkResult.fine === undefined) ? checkResult.fine : (checkResult.unlock ? 0 : undefined));
        if (!checkResult.unlock)
            return;
    }
    // Level abschließen
    this.startNextLevel();
}

/**
* Benutzer über Misserfolg informieren
* @return undefined
*/
HumanDbmsTaskUI.prototype._tellUser = function(
    msg,        // Message to display
    fine        // Zu vergebene Anzahl an Fehlerpunkten
) {
    this.addMsg(msg);
    const level = this.getLevelInfo();
    if (fine !== undefined)
        this.fineCount += Number.isFinite(fine) ? fine : 1;
    else if (level.fine)
        this.fineCount += Number.isFinite(level.fine) ? level.fine : 1;
}

/**
* Das nächste Level ggf. freischalten und starten
* @return undefined
*/
HumanDbmsTaskUI.prototype.startNextLevel = function() {
    const countDone = this._levelMgr.active + ((this._levelMgr.active + 1 == this.task.levels.length) ? 2 : 1);
    if (countDone > this._levelMgr.done) {
        this._levelMgr.done = countDone;
        if ((typeof this.onlevelunlock == 'function') || (this.onlevelunlock instanceof Function))
            this.onlevelunlock.call(this, Math.min(countDone, this.task.levels.length));
    }
    ++this._levelMgr.active;
}

/**
* Event auslösen, wenn sich der Inhalt der eingegebenen Tabelle verändert hat
* @return undefined
*/
HumanDbmsTaskUI.prototype._procTableChange = function(csv) {
    if ((typeof this.ontablechange == 'function') || (this.ontablechange instanceof Function))
        this.ontablechange.call(this, this._editableTable.toString());
}
